#include "memory_incremental.h"

#include <stdlib.h>

/* Initialise the memory pool with an optional 'limit' in pages. */

MemoryIncremental::MemoryIncremental(unsigned int limit, offset_t region_size)
: Memory(region_size), _limit(limit)
{
    _limited = true;
}

MemoryIncremental::MemoryIncremental()
: Memory()
{
    _limited = false;
}

/* Allocate a block of the given 'size'. */

Region *MemoryIncremental::allocate(offset_t size)
{
    /* Attempt to allocate aligned memory. */

    void *current;

    /* Make the size appropriate for the invocation. */

    size = round_multiple(size, PAGE_SIZE);

    if (posix_memalign(&current, size, size))
        return NULL;

    return new Region((offset_t) current, (offset_t) current + size);
}

/* Allocate a new region of the configured size rounded to the nearest page. */

Region *MemoryIncremental::region()
{
    return region(_region_size);
}

/* Allocate a new region of the given 'size' rounded to the nearest page. */

Region *MemoryIncremental::region(offset_t size)
{
    std::lock_guard<std::mutex> guard(_lock);

    offset_t rounded = round(size, PAGE_SIZE);
    offset_t pages = rounded / PAGE_SIZE;

    /* Check for sufficient pages. */

    if (!_limited || (_limit >= pages))
    {
        /* Attempt to allocate aligned memory. */

        Region *region = allocate(rounded);

        if (region == NULL)
            return NULL;

        if (_limited)
            _limit -= pages;

        return region;
    }

    /* Return no region without sufficient pages. */

    else
        return NULL;
}

/* Release the allocated 'region'. */

void MemoryIncremental::release(Region *region)
{
    std::lock_guard<std::mutex> guard(_lock);

    offset_t rounded = round(region->size(), PAGE_SIZE);
    offset_t pages = rounded / PAGE_SIZE;

    if (_limited)
        _limit += pages;

    delete region;
}

// vim: tabstop=4 expandtab shiftwidth=4
